/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */

package de.arimond.dt.operators.distribution;

import java.io.File;
import java.io.Serializable;
import java.util.List;

import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.ParameterTypeString;
import com.rapidminer.parameter.UndefinedParameterError;

import de.arimond.dt.TimeMeasurement;
import de.arimond.dt.operators.distribution.DistributionFactory.DistributionDataLayerType;
import de.arimond.dt.operators.distribution.DistributionFactory.DistributionType;
import de.arimond.dt.operators.distribution.gridgain.Grid;

/**
 * 
 * @author Alexander Arimond
 *
 * @param <I>
 * @param <S>
 * @param <M>
 * @param <O>
 */

public abstract class AbstractDistributableOperator <I extends Serializable, S extends Serializable, M extends Serializable, O extends Serializable> extends Operator{

	
	private static final String PARAMETER_MODE = "mode";
	private static final String STANDALONE_MODE = "STANDALONE";
	
	private static final String PARAMETER_DATALAYER = "data layer";
	
	public AbstractDistributableOperator(OperatorDescription description) {
		super(description);
	}

	protected abstract DistributionMapReduce <S, M, O> getMapReduceSpecification();
	
	public abstract List<S> split(I input);
	
	protected abstract I getDistributionInput() throws OperatorException;
	
	protected abstract void deliverDistributionOutput(O output);
	
	private Distribution <S,M,O> distribution;
	
	public void doWork() throws OperatorException{

		TimeMeasurement time = new TimeMeasurement();
		
		DistributionType type =  DistributionType.valueOf(this.getParameter(PARAMETER_MODE));
		DistributionDataLayerType layerType = DistributionDataLayerType.valueOf(this.getParameter(PARAMETER_DATALAYER));
		
		Grid.nr_threads = getParameterAsInt("threads");
		
		if (type.equals(DistributionType.GRIDGAIN) && layerType.equals(DistributionDataLayerType.COHERENCE))
			Grid.startGridCoherence();
		
		distribution = DistributionFactory.create(type, layerType, getMapReduceSpecification());
		
		try {
			I input = getDistributionInput();
			List <S> splits = split(input);
			distribution.setNumberMaps(splits.size());
			for (int i=0; i< splits.size(); i++){
				distribution.provideMapInputData(i, splits.get(i));
			}
			// FIXME did this cause jobs werent distributed correctly
			try {
				Thread.sleep(4000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			O output = startComputation();
			deliverDistributionOutput(output);
			distribution.clearData();
		} catch (DistributionException e) {
			throw new OperatorException("distribution failed.", e);
		} finally {
			if (type.equals(DistributionType.GRIDGAIN))
				Grid.stopGrid();
		}


		
		int threads = Grid.nr_threads;
		
		String message = getParameterAsString("messageForTimeMeasurement");
		int machines = Integer.parseInt(message.split(" ")[1]);
		if (type.equals(DistributionFactory.DistributionType.HADOOP)){
			threads = threads / machines;
		}
		
		time.appendToFile(new File("/home2/a_arimon/evaluation/SURF.txt"), type.toString() + " threads " +  threads + " machines "+ message + " splits " + 1500/distribution.numberMaps );
	}
	
	public O startComputation() throws DistributionException, UndefinedParameterError{
		try{
			return distribution.startComputation();
		} finally {
			//distribution.close();
		}
	}
	
	
	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		ParameterType type = new ParameterTypeCategory(PARAMETER_MODE,
				"standalone, gridgain or hadoop", 
				new String [] {STANDALONE_MODE, DistributionType.GRIDGAIN.toString(), DistributionType.HADOOP.toString()}, 
				0);
		types.add(type);
		ParameterType type1 = new ParameterTypeCategory(PARAMETER_DATALAYER,
				"which data layer to use when distributing with gridgain", 
				new String [] {DistributionDataLayerType.COHERENCE.toString()}, 
				0);
		types.add(type1);
		//FIXME delete this
		ParameterType type2 = new ParameterTypeString("messageForTimeMeasurement", "");
		types.add(type2);
		
		types.add(new ParameterTypeInt("threads", "threads",1, 16)); // FIXME
		return types;
	}
	
}
