/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.distribution.clustering;

import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.clustering.CentroidClusterModel;
import com.rapidminer.tools.math.similarity.DistanceMeasure;

import de.arimond.dt.TimeMeasurement;
import de.arimond.dt.operators.distribution.DistributionDataLayer;
import de.arimond.dt.operators.distribution.DistributionMapReduce;

public class KMeansMapReduceSpec extends
		DistributionMapReduce<ExampleSet, Assignments, KMeansResult> {

	/**
	 * @author Alexander Arimond
	 */
	private static final long serialVersionUID = 4621701781085083370L;

	private double[] getAsDoubleArray(Example example, Attributes attributes) {
		double[] values = new double[attributes.size()];
		int i = 0;
		for (Attribute attribute : attributes) {
			values[i] = example.getValue(attribute);
			i++;
		}
		return values;
	}

	@Override
	public Assignments map(DistributionDataLayer<ExampleSet> dataLayer,
			int mapNr) throws Exception {
		
		
		//System.out.println("map " + mapNr);
		
		TimeMeasurement time = new TimeMeasurement();
		time.start();
		
		
		TimeMeasurement loadingTime = new TimeMeasurement();
		loadingTime.start();
		
		ExampleSet exampleSet = dataLayer.getMapInputData(mapNr);
		CentroidClusterModel model = (CentroidClusterModel) dataLayer.getData("model");
		//int k = (Integer) dataLayer.getData("k");
		int k = (Integer) dataLayer.getData("k");
		DistanceMeasure measure = (DistanceMeasure) dataLayer.getData("distanceMeasure");
		Attributes attributes = (Attributes) dataLayer.getData("attributes");
		
		loadingTime.stop();
		loadingTime.print("LOADING TIME MAP " + mapNr + ": ");
		
		if (exampleSet == null){
			throw new NullPointerException("null pointer at key " + mapNr);
		}
		
		int [] centroidAssignments = new int[exampleSet.size()];
		
		double [][] centroidSums = new double[k][attributes.size()];
		int [] nrOfAssigns = new int[k];
		
		int i = 0;
		for (Example example : exampleSet) {
			double[] exampleValues = getAsDoubleArray(example, attributes);
			double nearestDistance = measure.calculateDistance(model.getCentroidCoordinates(0), exampleValues);
			int nearestIndex = 0;
			for (int centroidIndex = 1; centroidIndex < k; centroidIndex++) {
				double distance = measure.calculateDistance(model.getCentroidCoordinates(centroidIndex), exampleValues);
				if (distance < nearestDistance) {
					nearestDistance = distance;
					nearestIndex = centroidIndex;
				}
			}
			centroidAssignments[i] = nearestIndex;
			//model.getCentroid(nearestIndex).assignExample(exampleValues);
			for (int d =0 ; d < exampleValues.length; d++)
				centroidSums[nearestIndex][d] += exampleValues[d];
			nrOfAssigns[nearestIndex]++;
			i++;
		}
		
		time.stop();
		time.print("MAP " + mapNr + ": ");
		
		return new Assignments(mapNr, centroidAssignments, centroidSums, nrOfAssigns);
	}

	@Override
	public KMeansResult reduce(List<Assignments> mapResults,
			DistributionDataLayer<ExampleSet> dataLayer) throws Exception {
		
		//System.out.println("reduce");
		
		TimeMeasurement time = new TimeMeasurement();
		time.start();
		
		PriorityQueue<Assignments> allAssigns = new PriorityQueue<Assignments>(mapResults.size(), new Comparator<Assignments>(){
			@Override
			public int compare(Assignments arg0, Assignments arg1) {
				return arg0.getSplitNr() - arg0.getSplitNr();
			}
			
		});
		
		int k = (Integer) dataLayer.getData("k");
		CentroidClusterModel model = (CentroidClusterModel) dataLayer.getData("model");
		
		// order the results of the map stage 
		int size = 0;
		for (Assignments assigns : mapResults){
			allAssigns.add(assigns);
			size += assigns.getCentroidAssignments().length;
		}
		
		int [] allCentroidAssignments = new int[size];
		
		int index  =0;
		for (Assignments assigns : allAssigns){
			final int [] centroidAssigns = assigns.getCentroidAssignments();
			// collect all assignments in a single array
			for (int i = 0; i< centroidAssigns.length; i++){
					allCentroidAssignments[index + i] = centroidAssigns[i];
			}
			index += centroidAssigns.length;
			
			// collect the new assignments from all maps within the model 
			for (int j = 0 ; j < k; j++){
				model.getCentroid(j).assignManyExamples(assigns.getCentroidSums()[j], assigns.getNrOfAssigns()[j]);
			}
		}
		
		time.stop();
		time.print("REDUCE: ");
		
		return new KMeansResult(model, allCentroidAssignments);
	}

}
