/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.distribution.clustering2;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.AttributeRole;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.Tools;
import com.rapidminer.example.set.SplittedExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.MemoryExampleTable;
import com.rapidminer.operator.OperatorCapability;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.clustering.CentroidClusterModel;
import com.rapidminer.operator.clustering.ClusterModel;
import com.rapidminer.operator.clustering.clusterer.DistanceMatrix;
import com.rapidminer.operator.clustering.clusterer.RMAbstractClusterer;
import com.rapidminer.operator.learner.CapabilityProvider;
import com.rapidminer.operator.ports.metadata.CapabilityPrecondition;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.tools.Ontology;
import com.rapidminer.tools.RandomGenerator;
import com.rapidminer.tools.math.similarity.DistanceMeasure;
import com.rapidminer.tools.math.similarity.numerical.EuclideanDistance;

import de.arimond.dt.TimeMeasurement;
import de.arimond.dt.operators.distribution.Distribution;
import de.arimond.dt.operators.distribution.DistributionException;
import de.arimond.dt.operators.distribution.DistributionFactory;
import de.arimond.dt.operators.distribution.DistributionFactory.DistributionDataLayerType;
import de.arimond.dt.operators.distribution.DistributionFactory.DistributionType;
import de.arimond.dt.operators.distribution.gridgain.Grid;


/**
 * This operator represents an implementation of k-means. This operator will create a cluster attribute if not present
 * yet.
 * 
 * The implementation is according to paper of C. Elkan:
 * - Using the Triangle Inequality to Accelerate k-Means -
 * Proceedings of the Twentieth International Conference on Machine Learning (ICML-2003), Washington DC, 2003
 * 
 * @author Alexander Arimond
 */

public class DistributionFastKMeans extends RMAbstractClusterer implements CapabilityProvider{
	/** The parameter name for &quot;the maximal number of clusters&quot; */
	public static final String PARAMETER_K = "k";

	/**
	 * The parameter name for &quot;the maximal number of runs of the k method with random initialization that are
	 * performed&quot;
	 */
	public static final String PARAMETER_MAX_RUNS = "max_runs";

	/** The parameter name for &quot;the maximal number of iterations performed for one run of the k method&quot; */
	public static final String PARAMETER_MAX_OPTIMIZATION_STEPS = "max_optimization_steps";

	private static final String PARAMETER_SPLITS = "splits";

	public DistributionFastKMeans(OperatorDescription description) {
		super(description);
		
		getExampleSetInputPort().addPrecondition(new CapabilityPrecondition(this, getExampleSetInputPort()));
	}

	// make a new ExampleSet which only contains the data of current subset
	public static ExampleSet cloneSubset(SplittedExampleSet set, int nr) {
		
		set.selectSingleSubset(nr);
		
//		// old table
//		ExampleTable oldTable = set.getExampleTable();
//		
//		// old attributes
//		Iterator <AttributeRole> all = set.getAttributes().allAttributeRoles();
//		
//		// copied attributes and specials mapping
//		LinkedList <Attribute> allAttrs = new LinkedList<Attribute>();
//		HashMap <Attribute, String> specialsMapping = new HashMap <Attribute, String>(); 
//		while (all.hasNext()){
//			AttributeRole ar = all.next();
//			Attribute a = ar.getAttribute();
//			Attribute copyOfA = AttributeFactory.createAttribute(a.getName(), a.getValueType());
//			copyOfA.setBlockType(a.getBlockType());
//			if (a.isNominal())
//				copyOfA.setMapping(a.getMapping());
//			copyOfA.setConstruction(a.getConstruction());
//			copyOfA.setDefault(a.getDefault());
//			copyOfA.setTableIndex(a.getTableIndex());
//			if (ar.isSpecial()){
//				specialsMapping.put(copyOfA, ar.getSpecialName());
//			}
//			allAttrs.add(copyOfA);
//		}
//
//		// new example table with copied attributes
//		MemoryExampleTable table = new MemoryExampleTable(allAttrs);
//
//		// copy all data rows
//		for (int i  = 0; i < set.size(); i++){
//			DataRow dataRow = set.getExample(i).getDataRow();
//            double [] newDataRowData = new double[oldTable.getNumberOfAttributes()];
//			for (int a = 0; a < oldTable.getNumberOfAttributes(); a++) {
//                Attribute attribute = oldTable.getAttribute(a);
//                if (attribute != null) {
//                    newDataRowData[a] = dataRow.get(attribute);
//                } else {
//                    newDataRowData[a] = Double.NaN;
//                }
//            }
//			table.addDataRow(new DoubleArrayDataRow(newDataRowData));
//		}
//		
//		// create example set and return
//		final ExampleSet resultSet = table.createExampleSet(specialsMapping);
//		return resultSet;
//
		Iterator <AttributeRole> all = set.getAttributes().allAttributeRoles();
		
		// FIXME this method does not "copy" the examples, the corresponding data rows are still linked to the old attributes
		// example.getValue(attr) therefore will only work on the original attributes
		
		// copied attributes and specials mapping
		LinkedList <Attribute> allAttrs = new LinkedList<Attribute>();
		HashMap <Attribute, String> specialsMapping = new HashMap <Attribute, String>(); 
		while (all.hasNext()){
			AttributeRole ar = all.next();
			Attribute a = ar.getAttribute();
			Attribute copyOfA = AttributeFactory.createAttribute(a.getName(), a.getValueType());
			copyOfA.setBlockType(a.getBlockType());
			if (a.isNominal())
				copyOfA.setMapping(a.getMapping());
			copyOfA.setConstruction(a.getConstruction());
			copyOfA.setDefault(a.getDefault());
			copyOfA.setTableIndex(a.getTableIndex());
			if (ar.isSpecial()){
				specialsMapping.put(copyOfA, ar.getSpecialName());
			}
			allAttrs.add(copyOfA);
		}
	
		MemoryExampleTable table = new MemoryExampleTable(allAttrs);
		for (Example example : set){
			table.addDataRow(example.getDataRow());
		}
		return table.createExampleSet(specialsMapping);
	}
	
	
	@Override
	public ClusterModel generateClusterModel(ExampleSet exampleSet) throws OperatorException {
		
		Distribution <ExampleSet, Assignments, Boolean> distribution = null;
		
		try{
			Grid.startGridCoherence();

			int k = getParameterAsInt(PARAMETER_K);
			int maxOptimizationSteps = getParameterAsInt(PARAMETER_MAX_OPTIMIZATION_STEPS);
			int maxRuns = getParameterAsInt(PARAMETER_MAX_RUNS);
			DistanceMeasure measure = new EuclideanDistance();
			measure.init(exampleSet);

			// checking and creating ids if necessary
			Tools.checkAndCreateIds(exampleSet);

			// additional checks
			Tools.onlyNonMissingValues(exampleSet, "KMeans");
			if (exampleSet.size() < k) {
				throw new UserError(this, 142, k);
			}

			// extracting attribute names
			Attributes attributes = exampleSet.getAttributes();
			ArrayList<String> attributeNames = new ArrayList<String>(attributes.size());
			for (Attribute attribute : attributes)
				attributeNames.add(attribute.getName());

			RandomGenerator generator = RandomGenerator.getRandomGenerator(this);
			double minimalIntraClusterDistance = Double.POSITIVE_INFINITY;
			CentroidClusterModel bestModel = null;
			int[] bestAssignments = null;


			final int nr_splits = getParameterAsInt(PARAMETER_SPLITS);
			distribution = DistributionFactory.create(DistributionType.GRIDGAIN, DistributionDataLayerType.COHERENCE, new FastKMeansInitSpec());
			distribution.setNumberMaps(nr_splits);

			SplittedExampleSet splittedES = new SplittedExampleSet(exampleSet, nr_splits, SplittedExampleSet.LINEAR_SAMPLING, true, 1);
			for (int i = 0; i < splittedES.getNumberOfSubsets(); i++){
				ExampleSet copy = cloneSubset(splittedES, i);
				final SplitInformation split = new SplitInformation(copy, i, k);
				distribution.provideMapInputData(i, copy);
				distribution.provideMapData(i, "split_data", split);
			}
			distribution.provideStaticData("attributes", attributes);
			distribution.provideStaticData("distanceMeasure", measure);
			//distribution.provideStaticData("clusterDistances", cDistances);

			for (int iter = 0; iter < maxRuns; iter++) {

				checkForStop();
				CentroidClusterModel model = new CentroidClusterModel(exampleSet, k, attributeNames, measure, getParameterAsBoolean(RMAbstractClusterer.PARAMETER_ADD_AS_LABEL), getParameterAsBoolean(RMAbstractClusterer.PARAMETER_REMOVE_UNLABELED));

				// init centroids by assigning one single, unique example!
				int i = 0;
				for (Integer index : generator.nextIntSetWithRange(0, exampleSet.size(), k)) {
					model.assignExample(i, getAsDoubleArray(exampleSet.getExample(index), attributes));
					i++;
				}
				model.finishAssign();
				distribution.provideStaticData("model", model);
				ClusterDistances cDistances = new ClusterDistances(k);
				//cDistances = (ClusterDistances)distribution.getStaticData("clusterDistances");
				computeClusterDistances(cDistances.matrix, cDistances.s, model, measure);
				distribution.provideStaticData("clusterDistances", cDistances);

				final int[] centroidAssignments = new int[exampleSet.size()];

				// initializing
				distribution.startComputation();

				distribution.setMapReduceSpecification(new FastKMeansOptimizingSpec());

				// optimization steps (repeat until convergence)
				boolean stable = false;
				for (int step = 0; (step < maxOptimizationSteps) && !stable; step++) {


					// step 1.
					cDistances = (ClusterDistances)distribution.getStaticData("clusterDistances");
					model = (CentroidClusterModel) distribution.getStaticData("model");
					computeClusterDistances(cDistances.matrix, cDistances.s, model, measure);
					distribution.provideStaticData("clusterDistances", cDistances);

					// step 2,3,4,5,6
					
					TimeMeasurement t = new TimeMeasurement();
					t.start();
					stable = distribution.startComputation();
					t.print("optimizing time: ");
					
					
					if (stable) break;


				}
				// assessing quality of this model
				double distanceSum = 0;
				i = 0;
				for (Example example : exampleSet) {
					double distance = measure.calculateDistance(model.getCentroidCoordinates(centroidAssignments[i]), getAsDoubleArray(example, attributes));
					distanceSum += distance * distance;
					i++;
				}
				if (distanceSum < minimalIntraClusterDistance) {
					model = (CentroidClusterModel)distribution.getStaticData("model");
					receiveCentroidAssignments(centroidAssignments, distribution);

					bestModel = model;
					minimalIntraClusterDistance = distanceSum;
					bestAssignments = centroidAssignments;
				}
			}
			bestModel.setClusterAssignments(bestAssignments, exampleSet);

			if (addsClusterAttribute()) {
				Attribute cluster = AttributeFactory.createAttribute("cluster", Ontology.NOMINAL);
				exampleSet.getExampleTable().addAttribute(cluster);
				exampleSet.getAttributes().setCluster(cluster);
				int i = 0;
				for (Example example : exampleSet) {
					example.setValue(cluster, "cluster_" + bestAssignments[i]);
					i++;
				}
			}

			return bestModel;
		} catch (DistributionException e){
			throw new OperatorException("distribution failed for some reason", e);
		} finally {
			if (distribution != null) distribution.clearData();
			Grid.stopGrid();
			
		}
	}

	private void receiveCentroidAssignments(int [] centroidAssignments,
			Distribution<ExampleSet, Assignments, Boolean> distribution) {
		int x = 0;
		for (int split = 0; split < distribution.getNumberMaps(); split ++){
			// FIXME think this will be slow
			final int [] c = ((SplitInformation)distribution.getMapData(split, "split_data")).c; 
			for ( int j = 0; j < c.length; j++){
				centroidAssignments[x] = c[j];
				x++;
			}
		}
	}

	// this is for step 1 of the paper algorithm
	private void computeClusterDistances(DistanceMatrix centroidDistances, double[] s, 
			CentroidClusterModel model,  DistanceMeasure measure) {
		for (int i = 0; i < model.getNumberOfClusters(); i++){
			s[i] = Double.POSITIVE_INFINITY;
		}
		for (int i = 0; i < model.getNumberOfClusters(); i++){
			for (int j = i+1; j < model.getNumberOfClusters(); j++){
				final double d = measure.calculateDistance(model.getCentroidCoordinates(i), model.getCentroidCoordinates(j));
				if (d < s[i]){
					s[i] = d;
				}
				if (d < s[j]){
					s[j] = d;
				}
				centroidDistances.set(i, j, d);
			}
		}
		for (int i = 0; i < model.getNumberOfClusters(); i++){
			s[i] = 0.5 * s[i];
		}
	}

	private double[] getAsDoubleArray(Example example, Attributes attributes) {
		double[] values = new double[attributes.size()];
		int i = 0;
		for (Attribute attribute : attributes) {
			values[i] = example.getValue(attribute);
			i++;
		}
		return values;
	}

	@Override
	public Class<? extends ClusterModel> getClusterModelClass() {
		return CentroidClusterModel.class;
	}

	@Override
	public boolean supportsCapability(OperatorCapability capability) {
		switch (capability) {
		case BINOMINAL_ATTRIBUTES:
		case POLYNOMINAL_ATTRIBUTES:
			return false;
		default:
			return true;
		}
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.add(new ParameterTypeInt(PARAMETER_K, "The number of clusters which should be detected.", 2, Integer.MAX_VALUE, 2, false));
		types.add(new ParameterTypeInt(PARAMETER_MAX_RUNS, "The maximal number of runs of k-Means with random initialization that are performed.", 1, Integer.MAX_VALUE, 10, false));
		types.add(new ParameterTypeInt(PARAMETER_MAX_OPTIMIZATION_STEPS, "The maximal number of iterations performed for one run of k-Means.", 1, Integer.MAX_VALUE, 100, false));
		
		types.add(new ParameterTypeInt(PARAMETER_SPLITS, "splits", 2, 100, 20));
		
		types.addAll(RandomGenerator.getRandomGeneratorParameters(this));
		return types;
	}
}
