/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.distribution.clustering2;

import java.util.Comparator;
import java.util.List;
import java.util.PriorityQueue;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.clustering.CentroidClusterModel;
import com.rapidminer.tools.math.similarity.DistanceMeasure;

import de.arimond.dt.TimeMeasurement;
import de.arimond.dt.operators.distribution.DistributionDataLayer;
import de.arimond.dt.operators.distribution.DistributionMapReduce;

public class FastKMeansOptimizingSpec extends
		DistributionMapReduce<ExampleSet, Assignments, Boolean> {

	/**
	 * 
	 */
	private static final long serialVersionUID = -1391653673334131516L;

	private double[] getAsDoubleArray(Example example, Attributes attributes) {
		double[] values = new double[attributes.size()];
		int i = 0;
		for (Attribute attribute : attributes) {
			values[i] = example.getValue(attribute);
			i++;
		}
		return values;
	}
	
	
	@Override
	public Assignments map(DistributionDataLayer<ExampleSet> dataLayer,
			int mapNr) throws Exception {
		
		TimeMeasurement t = new TimeMeasurement();
		t.start();
		SplitInformation splitInfo = (SplitInformation) dataLayer.getMapData(mapNr, "split_data");
		ExampleSet exampleSet = (ExampleSet) dataLayer.getMapInputData(mapNr);
		t.print("example set and split loading time: ");
		
		DistanceMeasure measure = (DistanceMeasure) dataLayer.getData("distanceMeasure");
		Attributes attributes = (Attributes) dataLayer.getData("attributes");

		t.start();
		
		//split.initializeFields();
		final double [][] l = splitInfo.l;
		final double [] u = splitInfo.u;
		final boolean [] r = splitInfo.r;
		final int [] centroidAssignments = splitInfo.c;
		
		
		CentroidClusterModel model =(CentroidClusterModel) dataLayer.getData("model");
		
		/////////////////////
		// recalculate bounds before optimizing, by this we save one mapreduce operation
		
		double [] mean_distances = (double []) dataLayer.getData("mean_distances");
		if (mean_distances != null){ // do this only if mean_distances == null, i.e. not directly after initialization

			final int k = splitInfo.l[0].length;
			// step 5 & 6
			int x = 0;
			for (x=0; x < splitInfo.getSize(); x++){
				// step 5
				for (int c = 0; c < k; c++){
					final double d = l[x][c] - mean_distances[c]; 
					if (d > 0)
						l[x][c] = d;
					else
						l[x][c] = 0;
				}
				// step 6
				u[x] = u[x] + mean_distances[centroidAssignments[x]];
				r[x] = true;
			}
		}
		
		////////////////////////
		

		ClusterDistances cDistances = (ClusterDistances) dataLayer.getData("clusterDistances");

		final double [] s = cDistances.s;
		
		
		final double [][] centroidSums  = new double [model.getNumberOfClusters()][attributes.size()];
		final int [] nrOfAssigns = new int [model.getNumberOfClusters()];
		
		int x = 0;
		for (Example example : exampleSet) {
			final double [] exampleValue = getAsDoubleArray(example, attributes);

			// step 2.
			if ( u[x] <= s[centroidAssignments[x]]){
				// nothing
			} else {
				// step 3.
				for (int c = 0; c < model.getNumberOfClusters(); c ++){
					if (c != centroidAssignments[x]  // (i)
					    && u[x] > l[x][c] 			// (ii)
					    && u[x] > 0.5 * cDistances.matrix.get(centroidAssignments[x], c) // (iii)
					    ){
						//step 3a.
						final double d_x_c;   // d(x,c(x))
						if (r[x]){
							d_x_c = measure.calculateDistance(exampleValue, model.getCentroidCoordinates(centroidAssignments[x]));
							l[x][centroidAssignments[x]] = d_x_c;
							u[x] = d_x_c;
							r[x] = false;
						} else {
							d_x_c = u[x];
						}
						// step 3b.
						if (d_x_c > l[x][c] && d_x_c > 0.5 * cDistances.matrix.get(centroidAssignments[x], c)){
							final double d_x_c_new = measure.calculateDistance(exampleValue, model.getCentroidCoordinates(c)); // d(x,c)
							l[x][c] = d_x_c_new;
							if (d_x_c_new < d_x_c) {
								centroidAssignments[x] = c;
								u[x] = d_x_c_new;
							}
						}
					}
				}

			}
			//model.assignExample(centroidAssignments[x], exampleValue);
			for (int d = 0; d < model.getAttributeNames().length; d ++){
				centroidSums[centroidAssignments[x]][d] += exampleValue[d];
			}
			nrOfAssigns[centroidAssignments[x]] ++;
			x++;
		}
		t.print("map time: ");
		t.start();
		dataLayer.putMapData(mapNr, "split_data", splitInfo);
		t.print("store split again: ");
		return new Assignments(splitInfo.getNr(), centroidSums, nrOfAssigns);
	}

	@Override
	public Boolean reduce(List<Assignments> mapResults,
			DistributionDataLayer<ExampleSet> dataLayer) throws Exception {
		System.out.println("calculating new means (reduce)");

		TimeMeasurement t = new TimeMeasurement();
		t.start();

		PriorityQueue<Assignments> allAssigns = new PriorityQueue<Assignments>(mapResults.size(), new Comparator<Assignments>(){
			@Override
			public int compare(Assignments arg0, Assignments arg1) {
				return arg0.getSplitNr() - arg0.getSplitNr();
			}

		});

		// sort assignments due to split nr
		for (Assignments assigns : mapResults){
			allAssigns.add(assigns);
		}

		CentroidClusterModel model = (CentroidClusterModel) dataLayer.getData("model");

		for (Assignments assigns : allAssigns){
			for (int c = 0; c < model.getNumberOfClusters(); c++){
				model.getCentroid(c).assignManyExamples(assigns.getCentroidSums()[c], assigns.getNrOfAssigns()[c]);
			}
		}
		
		// step 4
		final double [][] m_old = new double [model.getNumberOfClusters()][model.getAttributeNames().length]; // needed for step 4
		// first store old c
		for (int c=0; c < model.getNumberOfClusters() ; c++){
			m_old[c] = model.getCentroidCoordinates(c);
		}
		
		// then compute the m(c) - here this is same as step 7
		boolean stable = model.finishAssign();
		if (stable){
			dataLayer.putData("model", model);
			return new Boolean(stable);
		}

		
		// compute all d(c,m(c))
		final DistanceMeasure measure = (DistanceMeasure) dataLayer.getData("distanceMeasure");
		final double [] mean_distances = new double [model.getNumberOfClusters()];
		for (int c = 0; c < model.getNumberOfClusters(); c++){
			mean_distances [c] = measure.calculateDistance(m_old[c], model.getCentroidCoordinates(c));
		}

		dataLayer.putData("mean_distances", mean_distances);
		dataLayer.putData("model", model);
		t.print("reduce time: ");
		return new Boolean(stable);
	}

}
