/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.distribution.clustering2;

import java.util.Map;
import java.util.Set;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.clustering.CentroidClusterModel;
import com.rapidminer.tools.math.similarity.DistanceMeasure;
import com.tangosol.util.InvocableMap.Entry;
import com.tangosol.util.InvocableMap.EntryProcessor;

import de.arimond.dt.TimeMeasurement;

/**
 * 
 * @author Alexander Arimond
 *
 */
public class OptimizingEntryProcessor implements EntryProcessor {

	/**
	 * 
	 */
	private static final long serialVersionUID = -7625622546569646168L;
	private ExampleSet exampleSet;
	private DistanceMeasure measure;
	private Attributes attributes;
	private CentroidClusterModel model;
	private double[] mean_distances;
	private ClusterDistances cDistances;

	public OptimizingEntryProcessor(ExampleSet exampleSet, DistanceMeasure measure, Attributes attributes, CentroidClusterModel model, double[] meanDistances, ClusterDistances cDistances){
		this.exampleSet = exampleSet;
		this.measure = measure;
		this.attributes = attributes;
		this.model = model;
		this.mean_distances = meanDistances;
		this.cDistances = cDistances;
	}
	
	@Override
	public Object process(Entry entry) {

		SplitInformation splitInfo = (SplitInformation)entry.getValue();
		
		TimeMeasurement t = new TimeMeasurement();
		t.start();
		
		//split.initializeFields();
		final double [][] l = splitInfo.l;
		final double [] u = splitInfo.u;
		final boolean [] r = splitInfo.r;
		final int [] centroidAssignments = splitInfo.c;
				
		/////////////////////
		// recalculate bounds before optimizing, by this we save one mapreduce operation
		
		if (mean_distances != null){ // do this only if mean_distances == null, i.e. not directly after initialization

			System.out.println("recalculate bounds");
			
			final int k = splitInfo.l[0].length;
			// step 5 & 6
			int x = 0;
			for (x=0; x < splitInfo.getSize(); x++){
				// step 5
				for (int c = 0; c < k; c++){
					final double d = l[x][c] - mean_distances[c]; 
					if (d > 0)
						l[x][c] = d;
					else
						l[x][c] = 0;
				}
				// step 6
				u[x] = u[x] + mean_distances[centroidAssignments[x]];
				r[x] = true;
			}
		}
		
		////////////////////////

		final double [] s = cDistances.s;
		
		
		final double [][] centroidSums  = new double [model.getNumberOfClusters()][attributes.size()];
		final int [] nrOfAssigns = new int [model.getNumberOfClusters()];
		
		int avoidedLooks = 0;
		int allLooks = 0;
		int x = 0;
		for (Example example : exampleSet) {
			final double [] exampleValue = getAsDoubleArray(example, attributes);

			// step 2.
			if ( u[x] <= s[centroidAssignments[x]]){
				// nothing
			} else {
				// step 3.
				for (int c = 0; c < model.getNumberOfClusters(); c ++){
					allLooks ++;
					if (c != centroidAssignments[x]  // (i)
					    && u[x] > l[x][c] 			// (ii)
					    && u[x] > 0.5 * cDistances.matrix.get(centroidAssignments[x], c) // (iii)
					    ){
						//step 3a.
						final double d_x_c;   // d(x,c(x))
						if (r[x]){
							d_x_c = measure.calculateDistance(exampleValue, model.getCentroidCoordinates(centroidAssignments[x]));
							l[x][centroidAssignments[x]] = d_x_c;
							u[x] = d_x_c;
							r[x] = false;
						} else {
							d_x_c = u[x];
						}
						// step 3b.
						if (d_x_c > l[x][c] && d_x_c > 0.5 * cDistances.matrix.get(centroidAssignments[x], c)){
							final double d_x_c_new = measure.calculateDistance(exampleValue, model.getCentroidCoordinates(c)); // d(x,c)
							l[x][c] = d_x_c_new;
							if (d_x_c_new < d_x_c) {
								centroidAssignments[x] = c;
								u[x] = d_x_c_new;
							}
						}
					} else {
						avoidedLooks ++;
					}
				}

			}
			//model.assignExample(centroidAssignments[x], exampleValue);
			for (int d = 0; d < model.getAttributeNames().length; d ++){
				centroidSums[centroidAssignments[x]][d] += exampleValue[d];
			}
			nrOfAssigns[centroidAssignments[x]] ++;
			x++;
		}
		t.print("avoided looks: " + avoidedLooks+ " of " + allLooks+ " map time: ");
		t.start();
		
		//dataLayer.putMapData(mapNr, "split_data", splitInfo);
		t.print("store split again: ");
		entry.setValue(splitInfo);
		
		return new Assignments(splitInfo.getNr(), centroidSums, nrOfAssigns);
	}

	private double[] getAsDoubleArray(Example example, Attributes attributes) {
		double[] values = new double[attributes.size()];
		int i = 0;
		for (Attribute attribute : attributes) {
			values[i] = example.getValue(attribute);
			i++;
		}
		return values;
	}
	
	@Override
	public Map processAll(Set setEntries) {

		// TODO Auto-generated method stub
		return null;
	}

}
