/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.distribution.clustering3;

import java.util.ArrayList;
import java.util.List;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.clustering.CentroidClusterModel;
import com.rapidminer.tools.RandomGenerator;
import com.rapidminer.tools.math.similarity.DistanceMeasure;

import de.arimond.dt.operators.distribution.DistributionDataLayer;
import de.arimond.dt.operators.distribution.DistributionMapReduce;

/**
 * 
 * @author Alexander Arimond
 *
 */
public class KMeansMapReduceSpec extends
		DistributionMapReduce<ExampleSet, KMeansResult, KMeansResult> {

	/**
	 * 
	 */
	private static final long serialVersionUID = 4621701781085083370L;

	private double[] getAsDoubleArray(Example example, Attributes attributes) {
		double[] values = new double[attributes.size()];
		int i = 0;
		for (Attribute attribute : attributes) {
			values[i] = example.getValue(attribute);
			i++;
		}
		return values;
	}

	@Override
	public KMeansResult map(DistributionDataLayer<ExampleSet> dataLayer,
			int mapNr) throws Exception {
		
		ExampleSet exampleSet = dataLayer.getMapInputData(mapNr);
		Long seed = (Long) dataLayer.getMapData(mapNr, "seed");
		int k = (Integer)dataLayer.getData("k");
		Attributes attributes = (Attributes) dataLayer.getData("attributes");
		ArrayList<String> attributeNames = (ArrayList <String>)dataLayer.getData("attributeNames");
		DistanceMeasure measure = (DistanceMeasure) dataLayer.getData("distanceMeasure");
		boolean add_as_label = (Boolean) dataLayer.getData("add_as_label");
		boolean remove_unlabeled = (Boolean) dataLayer.getData("remove_unlabeled");
		
		int maxOptimizationSteps = (Integer) dataLayer.getData("maxOptimizationSteps");
		
		RandomGenerator generator = new RandomGenerator(seed);

		CentroidClusterModel model = new CentroidClusterModel(exampleSet, k, attributeNames, measure, add_as_label, remove_unlabeled);
		
		// init centroids by assigning one single, unique example!
		int i = 0;
		for (Integer index : generator.nextIntSetWithRange(0, exampleSet.size(), k)) {
			model.assignExample(i, getAsDoubleArray(exampleSet.getExample(index), attributes));
			i++;
		}
		model.finishAssign();
		
		int[] centroidAssignments = new int[exampleSet.size()];
		boolean stable = false;
		for (int step = 0; (step < maxOptimizationSteps) && !stable; step++) {
			// assign examles to new centroids
			i = 0;
			for (Example example : exampleSet) {

				double[] exampleValues = getAsDoubleArray(example, attributes);
				double nearestDistance = measure.calculateDistance(model.getCentroidCoordinates(0), exampleValues);
				int nearestIndex = 0;
				for (int centroidIndex = 1; centroidIndex < k; centroidIndex++) {
					double distance = measure.calculateDistance(model.getCentroidCoordinates(centroidIndex), exampleValues);
					if (distance < nearestDistance) {
						nearestDistance = distance;
						nearestIndex = centroidIndex;
					}
				}
				centroidAssignments[i] = nearestIndex;
				model.getCentroid(nearestIndex).assignExample(exampleValues);
				i++;
			}
			// finishing assignment
			stable = model.finishAssign();
		}
		// assessing quality of this model
		double distanceSum = 0;
		i = 0;
		for (Example example : exampleSet) {
			double distance = measure.calculateDistance(model.getCentroidCoordinates(centroidAssignments[i]), getAsDoubleArray(example, attributes));
			distanceSum += distance * distance;
			i++;
		}
		
		return new KMeansResult(model, centroidAssignments, distanceSum);
	}

	@Override
	public KMeansResult reduce(List<KMeansResult> mapResults,
			DistributionDataLayer<ExampleSet> dataLayer) throws Exception {
		
		double minimal = Integer.MAX_VALUE;
		int index = 0;
		for (int i=0; i<mapResults.size(); i++){
			if (mapResults.get(i).getDistanceSum() < minimal){
				minimal = mapResults.get(i).getDistanceSum();
				index = i;
			}
		}
		return mapResults.get(index);
	}



}
