/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.features;

import java.util.List;
import java.util.Vector;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Iterator;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSetFactory;
import com.rapidminer.operator.AbstractExampleSetProcessing;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.metadata.AttributeMetaData;
import com.rapidminer.operator.ports.metadata.ExampleSetMetaData;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.parameter.UndefinedParameterError;
import com.rapidminer.tools.Ontology;


/**
 * This reads features from feature files like the Fire_histogram format
 * @author Alexander Arimond
 *
 */


public class ReadFeaturesOperator extends AbstractExampleSetProcessing {

	public ReadFeaturesOperator(OperatorDescription description) {
		super(description);
	}

	class FeaturesFileReader {
		private File featuresFile;
		private boolean hasRead = false;
		private String featureType;
		private Integer dim;
		private Integer counter;
		private Integer steps;
		private Integer min;
		private Integer max;
		private Double[] dataValues;
		
		public FeaturesFileReader(File featuresFile){
			this.featuresFile = featuresFile;
		}
		
		public void read() throws IOException{
			if(!hasRead){
				_read();
			}
		}
		
		private void _read() throws IOException{
			final BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(featuresFile)));
		    this.featureType = reader.readLine();
		    this.dim = new Integer(reader.readLine().split(" ")[1]);
		    this.counter = new Integer(reader.readLine().split(" ")[1]);
		    this.steps = new Integer(reader.readLine().split(" ")[1]);
		    this.min = new Integer(reader.readLine().split(" ")[1]);
		    this.max = new Integer(reader.readLine().split(" ")[1]);
		    final String [] dataValueStrings = reader.readLine().replaceFirst("data ", "").split(" ");
		    final Double [] dataValues = new Double[dataValueStrings.length];
		    for (int j=0; j<dataValueStrings.length; j++){
		    	dataValues[j] = new Double(dataValueStrings[j]);
		    }
			this.dataValues = dataValues;
			reader.close();
		}
		
		public String getFeatureType(){
			return this.featureType;
		}
		public Integer getDim(){
			return this.dim;
		}
		public Integer getCounter(){
			return this.counter;
		}
		public Integer getSteps(){
			return this.steps;
		}
		public Integer getMin(){
			return this.min;
		}
		public Integer getMax(){
			return this.max;
		}
		public Double [] getDataValues(){
			return this.dataValues;
		}
	}

	@Override
	public ExampleSet apply(ExampleSet exampleSet) throws OperatorException {
		final Iterator <Example> iter = exampleSet.iterator();
		final Attributes attrs = exampleSet.getAttributes();
		final Attribute pictureAttr = attrs.get("picture");
		final Attribute labelAttr = attrs.get("label");
		Double [] labels = new Double [exampleSet.size()];
		Object [][] data = new Object[exampleSet.size()][64 + 1];
		int i = 0;
		while (iter.hasNext()){
			final Example row = iter.next();
			final String filename = row.getNominalValue(pictureAttr);
			final String labelString = row.getNominalValue(labelAttr);
			final Double label = (labelString.equals("sailing")) ? 1.0 : 0.0;
			final File file = new File(filename);
			final File featuresFile = new File(file.getParent(), file.getName() + ".22concepts-plsa64.viswords.histo");
			final FeaturesFileReader reader = new FeaturesFileReader(featuresFile);
			try {
				reader.read();
			} catch (IOException e) {
				this.logError(e.getMessage());
				throw new OperatorException("IOException while reading feature files.", e);
			}
			labels[i] = label;
			final Double [] dataValues = reader.getDataValues();
			for (int j=0; j<dataValues.length ; j++){
				data[i][j+1] = dataValues[j];
			}
			data [i][0] = file.getAbsolutePath();
			i++;
		}
		final ExampleSet resultSet = ExampleSetFactory.createExampleSet(data, labels);
		resultSet.getAttributes().setLabel(resultSet.getAttributes().get("label"));
		resultSet.getAttributes().setId(resultSet.getAttributes().get("att1"));
		resultSet.getAttributes().get("att1").setName("filename");
		return resultSet;
	}

	@Override
	protected MetaData modifyMetaData(ExampleSetMetaData metaData)
			throws UndefinedParameterError {
		final ExampleSetMetaData md = new ExampleSetMetaData();
		// set label attribute meta data
		final List <AttributeMetaData> attrs = new Vector <AttributeMetaData>();
		final AttributeMetaData labelAttr = new AttributeMetaData("label", Ontology.BINOMINAL);
		labelAttr.setRole("label");
		// set id attribute meta data
	    final AttributeMetaData filename = new AttributeMetaData("filename", Ontology.NOMINAL);
	    filename.setRole("id");
	    attrs.add(filename);
	    // set data attributes metadata
		attrs.add(labelAttr);
		for (int i=1; i<=64 ; i++){
			attrs.add(new AttributeMetaData("att" + i, Ontology.NUMERICAL));
		}

		md.setAttributes(attrs);
		return md;
	}
}
