/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.features.extraction;

import java.io.File;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.DoubleArrayDataRow;
import com.rapidminer.example.table.MemoryExampleTable;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.ExampleSetMetaData;
import com.rapidminer.operator.ports.metadata.GenerateNewExampleSetMDRule;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.tools.Ontology;
import com.rapidminer.tools.att.AttributeSet;

import de.arimond.dt.TimeMeasurement;
import de.arimond.dt.features.Feature;
import de.arimond.dt.features.SURFFeatures;
import de.arimond.dt.features.SURFFeatures.SURFFeature;
import de.arimond.dt.features.extraction.SURFFeatureExtractor;
import de.arimond.dt.operators.distribution.images.FeatureExtractionException;
import de.arimond.dt.operators.images.ImageObject;
import de.arimond.dt.operators.images.ImageObjectList;
import de.arimond.dt.preprocessing.GrayScaleImageConverter;

/**
 * 
 * @author Alexander Arimond
 *
 */

public class SURFExtraction extends Operator {

	private final InputPort inputPort = getInputPorts().createPort("input", ImageObjectList.class);
	private final OutputPort outputPort = getOutputPorts().createPort("features");
	private final OutputPort original = getOutputPorts().createPort("original");
	
	public SURFExtraction(OperatorDescription description) {
		super(description);
		inputPort.addPrecondition(new SimplePrecondition(inputPort, new MetaData(ImageObjectList.class), true));
		GenerateNewExampleSetMDRule rule = new GenerateNewExampleSetMDRule(outputPort){
			public MetaData modifyMetaData(ExampleSetMetaData unmodifiedMD){
				return SURFExtraction.this.getResultMetaData(unmodifiedMD);
			}
		};
		getTransformer().addRule(rule);
		getTransformer().addPassThroughRule(inputPort, original);
	}

	@Override
	public void doWork() throws OperatorException{
		TimeMeasurement time = new TimeMeasurement();
		time.start();
		
		ImageObjectList images = inputPort.getData();
		LinkedList<SURFFeatures> feats = new LinkedList<SURFFeatures>();
		for (ImageObject image : images.getList()){
			try {
				feats.add(extractFeatures(image));
			} catch (FeatureExtractionException e) {
				throw new OperatorException("exception during feature extraction", e);
			}
		}
		outputPort.deliver(createExampleSet(feats));
		time.appendToFile(new File("/home2/a_arimon/evaluation/SURF_standalone.txt"), "");
	}
	
	public ExampleSet createExampleSet(List<SURFFeatures> results) {
		final int dim = 128;
		AttributeSet set = new AttributeSet();
        final Attribute idAttribute = AttributeFactory.createAttribute("filename", Ontology.NOMINAL);
        set.addAttribute(idAttribute);
        for (int i = 0; i < dim; i++) {
            set.addAttribute(AttributeFactory.createAttribute("att" + i, Ontology.NUMERICAL));
        }

        // create table
        final MemoryExampleTable table = new MemoryExampleTable(set.getAllAttributes());
        for (SURFFeatures features : results){
        	//F features = values.next();
        	Iterator <SURFFeature> iter = features.iterator();
        	while (iter.hasNext()){
        		Feature feat = iter.next();
	        	final double[] dataRow = new double[1 + dim];
	        	
	        	dataRow[0] = idAttribute.getMapping().mapString(features.getID());
	        	for (int j = 1; j < dim + 1; j++){
	        		dataRow[j] = feat.values[j -1];
	        	}
	        	table.addDataRow(new DoubleArrayDataRow(dataRow));
        	}
        }
        final ExampleSet resultSet = table.createExampleSet(set);
        resultSet.getAttributes().setId(idAttribute);
        return resultSet;
	}
	
	public static File workingDirectory = new File("/tmp/" +System.getProperty("user.name") + "_feature_extraction");
	
	
	protected SURFFeatures extractFeatures(ImageObject image) throws FeatureExtractionException {
		TimeMeasurement t = new TimeMeasurement();
		t.start();
		if (!workingDirectory.exists()){
			workingDirectory.mkdir();
		}
		final File imageFile = new File(image.getFilename());
		final File grayImage = new File(workingDirectory, imageFile.getName() + ".gray.pgm");
		try {
			try{
				// create gray image
				GrayScaleImageConverter.grayImage(imageFile, grayImage, "pgm");
			} catch (NoClassDefFoundError err){
				System.err.println("class not found, maybe JAI is not in classpath");
				throw err;
			}
			// extract features
			SURFFeatures feats = SURFFeatureExtractor.extractFeatures(grayImage, workingDirectory);
			feats.setID(image.getFilename());
			return feats;
		} catch (Exception e){
			throw new FeatureExtractionException(e);
		} finally {
			grayImage.delete();
			t.appendToFile(new File("/home2/a_arimon/evaluation/interestPointExtraction.txt"), "image");
		}
	}
	

	public MetaData getResultMetaData(ExampleSetMetaData newlyCreated) {
		// TODO
		return newlyCreated;
	}
	
}
