/*
 *  RapidMiner DisPaRe
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.arimond.dt.operators.features.extraction;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.DoubleArrayDataRow;
import com.rapidminer.example.table.MemoryExampleTable;
import com.rapidminer.operator.AbstractExampleSetProcessing;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.clustering.ClusterModel;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.metadata.ExampleSetMetaData;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.parameter.UndefinedParameterError;
import com.rapidminer.tools.Ontology;

/**
 * 
 * This operator takes a clustered set of patches and constructs a visual word 
 * for the picture to which the patches belong.
 * This visual word is acutally a histogram.
 * 
 * @author Alexander Arimond
 *
 */

public class VisualWords extends AbstractExampleSetProcessing {

	private final InputPort clusterModelInput = getInputPorts().createPort("cluster model");
	
	public VisualWords(OperatorDescription description) {
		super(description);
		clusterModelInput.addPrecondition(new SimplePrecondition(clusterModelInput, new MetaData(ClusterModel.class)));
	}	

	
	@Override
	public ExampleSet apply(ExampleSet exampleSet) throws OperatorException {
		
		ClusterModel model = clusterModelInput.getData();
		int nrOfClusters = model.getClusters().size();

		// count cluster assignments
		final Attribute imageAttr = exampleSet.getExampleTable().findAttribute("filename");
		final Attribute clusterAttr = exampleSet.getAttributes().getCluster();
		final HashMap <String, int []> counter = new HashMap <String, int []>();
		for (Example example : exampleSet){
			final String image   = example.getNominalValue(imageAttr);
			if (!counter.containsKey(image)){
				counter.put(image, new int[nrOfClusters + 1]);
			}
			final String cluster_name = example.getNominalValue(clusterAttr);
			final int cluster = Integer.parseInt(cluster_name.replaceFirst("cluster_", "")); //FIXME quite hacky
			counter.get(image)[cluster]++;
			counter.get(image)[nrOfClusters]++; // the last slot of the array counts the number of patches for this image
											    // which is needed for normalization later
		}

		// create new example set
		// create attributes
        final List<Attribute> attributeList = new ArrayList<Attribute>(nrOfClusters + 1);
        final Attribute idAttribute = AttributeFactory.createAttribute("filename", Ontology.NOMINAL);
        attributeList.add(idAttribute);
        for (int a = 0; a < nrOfClusters; a++) {
            attributeList.add(AttributeFactory.createAttribute(model.getCluster(a).toString(), Ontology.NUMERICAL));
        }

        // create table
        MemoryExampleTable table = new MemoryExampleTable(attributeList);
        for (String filename : counter.keySet()){
        	double[] dataRow = new double [attributeList.size() + 1];
        	dataRow [0] = idAttribute.getMapping().mapString(filename);
            for (int j = 1; j<attributeList.size(); j++){
            	int [] clusterCount = counter.get(filename);
            	dataRow [j] = ((double)clusterCount[j-1] / clusterCount[nrOfClusters]) * Math.log((double)counter.size() / model.getCluster(j-1).getNumberOfExamples());  
            	// normalization by tf-idf
            }
            table.addDataRow(new DoubleArrayDataRow(dataRow));
        }
        ExampleSet histogram = table.createExampleSet(idAttribute);
        histogram.getAttributes().setId(idAttribute);
        return histogram;
	}



	@Override
	protected ExampleSetMetaData getRequiredMetaData() {
		// TODO Auto-generated method stub
		return super.getRequiredMetaData();
	}



	@Override
	protected MetaData modifyMetaData(ExampleSetMetaData metaData)
			throws UndefinedParameterError {
		// TODO Auto-generated method stub
		return super.modifyMetaData(metaData);
	}

	
	
}
