package org.gridgain.examples.jbosscache;

import java.io.File;
import java.io.IOException;

import org.gridgain.grid.GridException;
import org.gridgain.grid.util.GridUtils;
import org.jboss.cache.Cache;
import org.jboss.cache.CacheFactory;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.parsing.XmlConfigurationParser;
import org.jgroups.JChannelFactory;

/**
 * 
 * @author Alexander Arimond
 *
 */

public final class GridJbossCacheManager {
    /** Cache configuration path relative to GridGain installation. */
    public static final String CACHE_CFG_PATH = "config/jbosscache/jboss-cache-new.xml";

    /** JGroups configuration path relative to GridGain installation. */
    public static final String JGROUPS_CFG_PATH = "config/jgroups/multicast/jgroups.xml";

    /** JBoss Cache instance. */
    private Cache<Object, Object> cache = null;

    /** Cache node for caching example data. */
    private Node<Object, Object> cacheRoot = null;

    /** Singleton instance. */
    private static GridJbossCacheManager instance = new GridJbossCacheManager();

    /**
     * Gets singleton.
     *
     * @return Singleton.
     */
    public static GridJbossCacheManager getInstance() {
        return instance;
    }

    /**
     * Ensure singleton.
     */
    private GridJbossCacheManager() {
        // No-op.
    }

    /**
     * Gets test data from data cache by ID.
     *
     * @param id Test data ID.
     * @return Cached data (possibly <tt>null</tt>).
     */
    public Object getSet(long id) {
        assert cacheRoot != null;
        
        Node<Object, Object> dataNode;
        // Try to get child node from cache.
        dataNode = cache.getNode(cacheRoot.getFqn().toString().concat("/" + id));

        // If data node does not exist, we lazily create it.
//        if (dataNode == null) {
//            dataNode = cache.getRoot().addChild(Fqn.fromString(cacheRoot.getFqn().toString().concat("/" + id)));
//        }

        assert dataNode != null;

        // Retrieve cached data from cache.
        return dataNode.get(id);
    }

    /**
     * Puts test data to the data cache.
     *
     * @param id index that associated with this test data.
     * @param data Data to store in cache.
     */
    public void putSet(long id, Object data) {
        assert cacheRoot != null;

        //Node<Object, Object> dataNode = cacheRoot.getChild(Fqn.fromList(Collections.singletonList(id)));

        Node<Object, Object> dataNode;
        // Try to get child node from cache.
        dataNode = cache.getNode(cacheRoot.getFqn().toString().concat("/" + id));
        
        // If data node does not exist, we lazily create it.
        if (dataNode == null) {
            dataNode = cache.getRoot().addChild(Fqn.fromString(cacheRoot.getFqn().toString().concat("/" + id)));
            //dataNode = cacheRoot.addChild(Fqn.fromList(Collections.singletonList(id)));
        }

        assert dataNode != null;

        // Cache data.
        dataNode.put(id, data);
    }
    
    public Object get(Object key){
    	return cache.getRoot().get(key);
    }

    public void put(Object key, Object value){
    	cache.getRoot().put(key, value);
    }

    /**
     * Starts JBoss Cache instance.
     * 
     * @throws GridException If start failed.
     */
    public void start() throws GridException {
        File cacheCfg = GridUtils.resolveGridGainPath(CACHE_CFG_PATH);

        if (cacheCfg == null) {
            throw new GridException("Failed to find cache configuration file: " + CACHE_CFG_PATH);
        }

        File jgroupsCfg = GridUtils.resolveGridGainPath(JGROUPS_CFG_PATH);

        if (jgroupsCfg == null) {
            throw new GridException("Failed to find jgroups configuration: " + JGROUPS_CFG_PATH);
        }

        // Make sure JBoss Cache and GridGain are on the same "wave length".
        JChannelFactory factory = new JChannelFactory();

        try {
            factory.setMultiplexerConfig(jgroupsCfg.getCanonicalPath());
        }
        catch (Exception e) {
            throw new GridException("Failed to start Data Manager.", e);
        }

        try {
            XmlConfigurationParser parser = new XmlConfigurationParser();

            // Start JBoss Cache cache with shared JGroups configuration.
            Configuration svrCfg = parser.parseFile(cacheCfg.getCanonicalPath());

            svrCfg.getRuntimeConfig().setMuxChannelFactory(factory);

            CacheFactory<Object, Object> dataFactory = new DefaultCacheFactory<Object, Object>();
            
            // Instantiate and start JBoss Cache.
            cache = dataFactory.createCache(svrCfg);
            
            // Cache node for storing example data.
            cacheRoot = cache.getRoot().addChild(Fqn.fromString("/gridgain/example"));

            //Region region = cache.getRegion(Fqn.fromString("/gridgain/example"), true);
            System.out.println("CLUSTER NAME: " + cache.getConfiguration().getClusterName());
            System.out.println("CACHE MODE: " + cache.getConfiguration().getCacheModeString());
            System.out.println("CACHELOADER: " + cache.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().getClassName());
            System.out.println("PROPERTIES: " + cache.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().getProperties().toString());
            
        }
        catch (IOException e) {
            throw new GridException("Failed to start Data Manager.", e);
        }

        System.out.println("JBoss Cache data manager started.");
    }

    public void setRegionClassLoader(){
        this.cache.getRegion(Fqn.fromString("/gridgain/example"), true).registerContextClassLoader(this.getClass().getClassLoader());
    }
    
    
    /**
     * Stops JBoss Cache instance.
     */
    public void stop() {
        if (cache != null) {
            cache.stop();
            cache.destroy();
        }

        System.out.println("JBoss Cache data manager stopped.");
    }

	public void clear() {
		cacheRoot.clearData();
	}
}
